@@ -25,27 +25,38 @@
  * @{
  */
 
-#ifndef _KNOTD_CONF_H_
-#define _KNOTD_CONF_H_
+#pragma once
 
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <stdbool.h>
 
 #include <urcu.h>
 
 #include "libknot/dname.h"
-#include "libknot/util/descriptor.h"
-#include "libknot/tsig.h"
-#include "common/lists.h"
+#include "libknot/rrtype/tsig.h"
+#include "libknot/dnssec/key.h"
+#include "libknot/dnssec/policy.h"
+#include "common-knot/lists.h"
 #include "common/log.h"
-#include "common/sockaddr.h"
-#include "common/general-tree.h"
+#include "knot/updates/acl.h"
+#include "common-knot/sockaddr.h"
+#include "common-knot/hattrie/hat-trie.h"
+#include "knot/nameserver/query_module.h"
 
 /* Constants. */
 #define CONFIG_DEFAULT_PORT 53
 #define CONFIG_NOTIFY_RETRIES 5  /*!< 5 retries (suggested in RFC1996) */
 #define CONFIG_NOTIFY_TIMEOUT 60 /*!< 60s (suggested in RFC1996) */
-#define CONFIG_DBSYNC_TIMEOUT (60*60) /*!< 1 hour. */
+#define CONFIG_DBSYNC_TIMEOUT 0 /*!< Sync immediately. */
+#define CONFIG_REPLY_WD 10 /*!< SOA/NOTIFY query timeout [s]. */
+#define CONFIG_HANDSHAKE_WD 5 /*!< [secs] for connection to make a request.*/
+#define CONFIG_IDLE_WD  20 /*!< [secs] of allowed inactivity between requests */
+#define CONFIG_MAXTCP 100  /*!< Default limit on incoming TCP clients. */
+#define CONFIG_RRL_SLIP 1 /*!< Default slip value. */
+#define CONFIG_RRL_SIZE 393241 /*!< Htable default size. */
+#define CONFIG_XFERS 10
+#define CONFIG_SERIAL_DEFAULT CONF_SERIAL_INCREMENT /*!< Default serial policy: increment. */
 
 /*!
  * \brief Configuration for the interface
@@ -55,14 +66,12 @@
  * listening and outgoing function.
  */
 typedef struct conf_iface_t {
-	node n;
-	char *name;       /*!< Internal name for the interface. */
-	char *address;    /*!< IP (IPv4/v6) address for this interface */
-	unsigned prefix;  /*!< IP subnet prefix. */
-	int port;         /*!< Port number for this interface */
-	int family;       /*!< Address family. */
-	knot_key_t *key;  /*!< TSIG key (only valid for remotes). */
-	sockaddr_t  via;  /*!< Used for remotes to specify qry endpoint.*/
+	node_t n;
+	char *name;                   /*!< Internal name for the interface. */
+	knot_tsig_key_t *key;         /*!< TSIG key (only applic for remotes). */
+	unsigned prefix;              /*!< IP subnet prefix (only applic for remotes). */
+	struct sockaddr_storage addr; /*!< Interface address. */
+	struct sockaddr_storage via;  /*!< Used for remotes to specify qry endpoint.*/
 } conf_iface_t;
 
 /*!
@@ -71,10 +80,29 @@ typedef struct conf_iface_t {
  * Used for zone ACL lists to prevent node duplication.
  */
 typedef struct conf_remote_t {
-	node n;               /*!< List node. */
+	node_t n;             /*!< List node. */
 	conf_iface_t *remote; /*!< Pointer to interface descriptor. */
 } conf_remote_t;
 
+/*!
+ * \brief Group of remotes list item.
+ *
+ * Holds the name of a remote in the list.
+ */
+typedef struct conf_group_remote_t {
+	node_t n;
+	char *name;
+} conf_group_remote_t;
+
+/*!
+ * \brief Group of remotes.
+ */
+typedef struct conf_group_t {
+	node_t n;	/*!< List node. */
+	char *name;	/*!< Unique name of the group. */
+	list_t remotes;	/*!< List of remote names. */
+} conf_group_t;
+
 /*!
  * \brief Zone configuration.
  *
@@ -85,32 +113,46 @@ typedef struct conf_remote_t {
  * zone transfers.  Same logic applies for the NOTIFY.
  */
 typedef struct conf_zone_t {
-	node n;
-	char *name;               /*!< Zone name. */
-	enum knot_rr_class cls;   /*!< Zone class (IN or CH). */
-	char *file;               /*!< Path to a zone file. */
-	char *db;                 /*!< Path to a database file. */
-	char *ixfr_db;            /*!< Path to a IXFR database file. */
-	size_t ixfr_fslimit;      /*!< File size limit for IXFR journal. */
-	int dbsync_timeout;       /*!< Interval between syncing to zonefile.*/
-	int enable_checks;        /*!< Semantic checks for parser.*/
-	int disable_any;          /*!< Disable ANY type queries for AA.*/
-	int notify_retries;       /*!< NOTIFY query retries. */
-	int notify_timeout;       /*!< Timeout for NOTIFY response (s). */
-	int build_diffs;          /*!< Calculate differences from changes. */
+	char *name;                /*!< Zone name. */
+	char *file;                /*!< Path to a zone file. */
+	char *storage;             /*!< Path to a storage dir. */
+	char *dnssec_keydir;       /*!< Path to a DNSSEC key dir. */
+	char *ixfr_db;             /*!< Path to a IXFR database file. */
+	int dnssec_enable;         /*!< DNSSEC: Online signing enabled. */
+	size_t ixfr_fslimit;       /*!< File size limit for IXFR journal. */
+	int sig_lifetime;          /*!< Validity period of DNSSEC signatures. */
+	int dbsync_timeout;        /*!< Interval between syncing to zonefile.*/
+	int enable_checks;         /*!< Semantic checks for parser.*/
+	int disable_any;           /*!< Disable ANY type queries for AA.*/
+	int notify_retries;        /*!< NOTIFY query retries. */
+	int notify_timeout;        /*!< Timeout for NOTIFY response (s). */
+	int build_diffs;           /*!< Calculate differences from changes. */
+	int serial_policy;         /*!< Serial policy when updating zone. */
 	struct {
-		list xfr_in;      /*!< Remotes accepted for for xfr-in.*/
-		list xfr_out;     /*!< Remotes accepted for xfr-out.*/
-		list notify_in;   /*!< Remotes accepted for notify-in.*/
-		list notify_out;  /*!< Remotes accepted for notify-out.*/
+		list_t xfr_in;     /*!< Remotes accepted for for xfr-in.*/
+		list_t xfr_out;    /*!< Remotes accepted for xfr-out.*/
+		list_t notify_in;  /*!< Remotes accepted for notify-in.*/
+		list_t notify_out; /*!< Remotes accepted for notify-out.*/
+		list_t update_in;  /*!< Remotes accepted for DDNS.*/
 	} acl;
+
+	struct query_plan *query_plan;
+	list_t query_modules;
 } conf_zone_t;
 
+/*!
+ * \brief Serial policy options.
+ */
+typedef enum conf_serial_policy_t {
+	CONF_SERIAL_INCREMENT	= 1 << 0,
+	CONF_SERIAL_UNIXTIME	= 1 << 1
+} conf_serial_policy_t;
+
 /*!
  * \brief Mapping of loglevels to message sources.
  */
 typedef struct conf_log_map_t {
-	node n;
+	node_t n;
 	int source; /*!< Log message source mask. */
 	int prios;  /*!< Log priorities mask. */
 } conf_log_map_t;
@@ -119,10 +161,10 @@ typedef struct conf_log_map_t {
  * \brief Log facility descriptor.
  */
 typedef struct conf_log_t {
-	node n;
+	node_t n;
 	logtype_t type;  /*!< Type of the log (SYSLOG/STDERR/FILE). */
 	char *file;      /*!< Filename in case of LOG_FILE, else NULL. */
-	list map;        /*!< Log levels mapping. */
+	list_t map;      /*!< Log levels mapping. */
 } conf_log_t;
 
 /*!
@@ -140,10 +182,19 @@ typedef enum conf_section_t {
  * \brief TSIG key list item.
  */
 typedef struct conf_key_t {
-	node n;
-	knot_key_t k;
+	node_t n;
+	knot_tsig_key_t k;
 } conf_key_t;
 
+/*!
+ * \brief Remote control interface.
+ */
+typedef struct conf_control_t {
+	conf_iface_t *iface; /*!< Remote control interface. */
+	list_t allow;        /*!< List of allowed remotes. */
+	bool have;           /*!< Set if configured. */
+} conf_control_t;
+
 /*!
  * \brief Main config structure.
  *
@@ -154,59 +205,80 @@ typedef struct conf_t {
 	 * System
 	 */
 	char *filename; /*!< Name of the config file. */
-	char *identity; /*!< Identity to return on CH TXT id.server. */
-	char *version;  /*!< Version for CH TXT version.{bind|server} */
-	char *storage;  /*!< Persistent storage path for databases and such. */
-	char *pidfile;  /*!< PID file path. */
+	char *identity; /*!< Identity to return on CH TXT id.server. or hostname.bind. */
+	char *version;  /*!< Version for CH TXT version.{bind|server}. */
+	char *rundir;   /*!< Run-time directory path. */
+	char *pidfile;  /*!< PID file location. */
 	char *nsid;     /*!< Server's NSID. */
 	size_t nsid_len;/*!< Server's NSID length. */
+	size_t max_udp_payload; /*!< Maximal UDP payload size. */
 	int   workers;  /*!< Number of workers per interface. */
+	int   bg_workers; /*!< Number of background workers. */
+	bool  async_start; /*!< Asynchronous startup. */
 	int   uid;      /*!< Specified user id. */
 	int   gid;      /*!< Specified group id. */
+	int   max_conn_idle; /*!< TCP idle timeout. */
+	int   max_conn_hs;   /*!< TCP of inactivity before first query. */
+	int   max_conn_reply; /*!< TCP/UDP query timeout. */
+	int   max_tcp_clients; /*!< TCP client limit. */
+	int    rrl;      /*!< Rate limit (in responses per second). */
+	size_t rrl_size; /*!< Rate limit htable size. */
+	int    rrl_slip;  /*!< Rate limit SLIP. */
+	int    xfers;     /*!< Number of parallel transfers. */
 
 	/*
 	 * Log
 	 */
-	list logs;        /*!< List of logging facilites. */
-	int logs_count;   /*!< Count of logging facilities. */
+	list_t logs;      /*!< List of logging facilites. */
 
 	/*
 	 * Interfaces
 	 */
-	list ifaces;      /*!< List of interfaces. */
-	int ifaces_count; /*!< Count of interfaces. */
+	list_t ifaces;    /*!< List of interfaces. */
 
 	/*
 	 * TSIG keys
 	 */
-	list keys;     /*!< List of TSIG keys. */
-	int key_count; /*!< Count of TSIG keys. */
+	list_t keys;   /*!< List of TSIG keys. */
 
 	/*
 	 * Remotes
 	 */
-	list remotes;     /*!< List of remotes. */
-	int remotes_count;/*!< Count of remotes. */
+	list_t remotes;    /*!< List of remotes. */
+
+	/*
+	 * Groups of remotes.
+	 */
+	list_t groups; /*!< List of groups of remotes. */
 
 	/*
 	 * Zones
 	 */
-	list zones;       /*!< List of zones. */
-	int zones_count;  /*!< Count of zones. */
-	int zone_checks;  /*!< Semantic checks for parser.*/
-	int disable_any;  /*!< Disable ANY type queries for AA.*/
-	int notify_retries; /*!< NOTIFY query retries. */
-	int notify_timeout; /*!< Timeout for NOTIFY response in seconds. */
-	int dbsync_timeout; /*!< Default interval between syncing to zonefile.*/
+	hattrie_t *zones;    /*!< List of zones. */
+	int zone_checks;     /*!< Semantic checks for parser.*/
+	int disable_any;     /*!< Disable ANY type queries for AA.*/
+	int notify_retries;  /*!< NOTIFY query retries. */
+	int notify_timeout;  /*!< Timeout for NOTIFY response in seconds. */
+	int dbsync_timeout;  /*!< Default interval between syncing to zonefile.*/
 	size_t ixfr_fslimit; /*!< File size limit for IXFR journal. */
 	int build_diffs;     /*!< Calculate differences from changes. */
-	general_tree_t *zone_tree; /*!< Zone tree for duplicate checking. */
-	
+	char *storage;       /*!< Storage dir. */
+	char *dnssec_keydir; /*!< DNSSEC: Path to key directory. */
+	int dnssec_enable;   /*!< DNSSEC: Online signing enabled. */
+	int sig_lifetime;    /*!< DNSSEC: Signature lifetime. */
+	int serial_policy;   /*!< Serial policy when updating zone. */
+	struct query_plan *query_plan;
+	list_t query_modules;
+
+	/*
+	 * Remote control interface.
+	 */
+	conf_control_t ctl;
+
 	/*
 	 * Implementation specifics
 	 */
-	list hooks;      /*!< List of config hooks. */
-	int hooks_count; /*!< Count of config hooks. */
+	list_t hooks;    /*!< List of config hooks. */
 	int _touched;    /*!< Bitmask of sections touched by last update. */
 } conf_t;
 
@@ -214,7 +286,7 @@ typedef struct conf_t {
  * \brief Config hook prototype.
  */
 typedef struct conf_hook_t {
-	node n;
+	node_t n;
 	int sections; /*!< Bitmask of watched sections. */
 	int (*update)(const conf_t*, void*); /*!< Function executed on config load. */
 	void *data;
@@ -231,7 +303,7 @@ typedef struct conf_hook_t {
  * \retval new structure if successful.
  * \retval NULL on error.
  */
-conf_t *conf_new(const char* path);
+conf_t *conf_new(char *path);
 
 /*!
  * \brief Register on-update callback.
@@ -247,18 +319,6 @@ conf_t *conf_new(const char* path);
 int conf_add_hook(conf_t * conf, int sections,
                   int (*on_update)(const conf_t*, void*), void *data);
 
-/*!
- * \brief Parse configuration from associated file.
- *
- * \note Registered callbacks may be executed if applicable.
- *
- * \param conf Configuration context.
- *
- * \retval KNOT_EOK on success.
- * \retval KNOT_EPARSEFAIL on parser error.
- */
-int conf_parse(conf_t *conf);
-
 /*!
  * \brief Parse configuration from string.
  *
@@ -294,13 +354,9 @@ void conf_free(conf_t *conf);
 /*!
  * \brief Find implicit configuration file.
  *
- * Ordering:
- * 1. ~/.knot/knot.conf (if exists)
- * 2. /etc/knot/knot.conf (fallback)
- *
  * \return Path to implicit configuration file.
  */
-char* conf_find_default();
+const char* conf_find_default();
 
 /*!
  * \brief Open singleton configuration from file.
@@ -331,17 +387,6 @@ static inline conf_t* conf() {
  * Utilities.
  */
 
-/*!
- * \brief Create new string from a concatenation of s1 and s2.
- *
- * \param s1 First string.
- * \param s2 Second string.
- *
- * \retval Newly allocated string on success.
- * \retval NULL on error.
- */
-char* strcdup(const char *s1, const char *s2);
-
 /*!
  * \brief Normalize file path and expand '~' placeholders.
  *
@@ -351,6 +396,18 @@ char* strcdup(const char *s1, const char *s2);
  */
 char* strcpath(char *path);
 
+/*! \brief Return the number of UDP threads according to the configuration. */
+size_t conf_udp_threads(const conf_t *conf);
+
+/*! \brief Return the number of TCP threads according to the configuration. */
+size_t conf_tcp_threads(const conf_t *conf);
+
+/*! \brief Return the number of background worker threads. */
+int conf_bg_threads(const conf_t *conf);
+
+/* \brief Initialize zone config. */
+void conf_init_zone(conf_zone_t *zone);
+
 /*! \brief Free zone config. */
 void conf_free_zone(conf_zone_t *zone);
 
@@ -360,9 +417,13 @@ void conf_free_key(conf_key_t *k);
 /*! \brief Free interface config. */
 void conf_free_iface(conf_iface_t *iface);
 
+/*! \brief Free remotes config. */
+void conf_free_remote(conf_remote_t *r);
+
+/*! \brief Free group config. */
+void conf_free_group(conf_group_t *group);
+
 /*! \brief Free log config. */
 void conf_free_log(conf_log_t *log);
 
-#endif /* _KNOTD_CONF_H_ */
-
 /*! @} */
