@@ -14,7 +14,6 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <config.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -27,28 +26,23 @@
 #endif
 
 #include "knot/knot.h"
-#include "common/descriptor.h"
+#include "common/mem.h"
+#include "libknot/descriptor.h"
 #include "knot/ctl/process.h"
 #include "knot/ctl/remote.h"
 #include "knot/conf/conf.h"
+#include "knot/zone/zonefile.h"
 #include "knot/zone/zone-load.h"
-#include "knot/server/socket.h"
 #include "knot/server/tcp-handler.h"
-#include "libknot/util/wire.h"
-#include "libknot/packet/query.h"
-#include "libknot/packet/response.h"
-#include "knot/zone/zone-load.h"
-#include "knot/zone/estimator.h"
+#include "libknot/packet/wire.h"
+#include "knot/ctl/estimator.h"
 
 /*! \brief Controller flags. */
 enum knotc_flag_t {
-	F_NULL = 0 << 0,
-	F_FORCE = 1 << 0,
-	F_VERBOSE = 1 << 1,
-	F_INTERACTIVE = 1 << 3,
-	F_UNPRIVILEGED = 1 << 5,
-	F_NOCONF = 1 << 6,
-	F_DRYRUN = 1 << 7
+	F_NULL =         0 << 0,
+	F_FORCE =        1 << 0,
+	F_VERBOSE =      1 << 1,
+	F_NOCONF =       1 << 2
 };
 
 /*! \brief Check if flag is present. */
@@ -84,15 +78,15 @@ static int cmd_signzone(int argc, char *argv[], unsigned flags);
 /*! \brief Table of remote commands. */
 knot_cmd_t knot_cmd_tbl[] = {
 	{&cmd_stop,       0, "stop",       "",       "\t\tStop server."},
-	{&cmd_reload,     0, "reload",     "",       "\tReload configuration and changed zones."},
-	{&cmd_refresh,    0, "refresh",    "[zone]", "\tRefresh slave zone (all if not specified)."},
-	{&cmd_flush,      0, "flush",      "",       "\t\tFlush journal and update zone files."},
+	{&cmd_reload,     0, "reload",     "<zone>", "\tReload configuration and changed zones."},
+	{&cmd_refresh,    0, "refresh",    "<zone>", "\tRefresh slave zone (all if not specified). Flag '-f' forces retransfer."},
+	{&cmd_flush,      0, "flush",      "<zone>", "\tFlush journal and update zone file (all if not specified)."},
 	{&cmd_status,     0, "status",     "",       "\tCheck if server is running."},
 	{&cmd_zonestatus, 0, "zonestatus", "",       "\tShow status of configured zones."},
 	{&cmd_checkconf,  1, "checkconf",  "",       "\tCheck current server configuration."},
-	{&cmd_checkzone,  1, "checkzone",  "[zone]", "Check zone (all if not specified)."},
-	{&cmd_memstats,   1, "memstats",   "[zone]", "Estimate memory use for zone (all if not specified)."},
-	{&cmd_signzone,   0, "signzone",   "[zone]", "Sign all zones with available DNSSEC keys."},
+	{&cmd_checkzone,  1, "checkzone",  "<zone>", "Check zone (all if not specified)."},
+	{&cmd_memstats,   1, "memstats",   "<zone>", "Estimate memory use for zone (all if not specified)."},
+	{&cmd_signzone,   0, "signzone",   "<zone>", "Sign all zones with available DNSSEC keys."},
 	{NULL, 0, NULL, NULL, NULL}
 };
 
@@ -109,7 +103,6 @@ void help(void)
 	       " -f, --force            \tForce operation - override some checks.\n"
 	       " -v, --verbose          \tVerbose mode - additional runtime information.\n"
 	       " -V, --version          \tPrint %s server version.\n"
-	       " -i, --interactive      \tInteractive mode (do not daemonize).\n"
 	       " -h, --help             \tPrint help and usage.\n",
 	       RUN_DIR "/knot.sock", PACKAGE_NAME);
 	printf("\nActions:\n");
@@ -123,11 +116,12 @@ void help(void)
 static int cmd_remote_print_reply(const knot_rrset_t *rr)
 {
 	/* Process first RRSet in data section. */
-	if (knot_rrset_type(rr) != KNOT_RRTYPE_TXT) {
+	if (rr->type != KNOT_RRTYPE_TXT) {
 		return KNOT_EMALF;
 	}
 
-	for (uint16_t i = 0; i < knot_rrset_rdata_rr_count(rr); i++) {
+	uint16_t rr_count = rr->rrs.rr_count;
+	for (uint16_t i = 0; i < rr_count; i++) {
 		/* Parse TXT. */
 		remote_print_txt(rr, i);
 	}
@@ -137,46 +131,46 @@ static int cmd_remote_print_reply(const knot_rrset_t *rr)
 
 static int cmd_remote_reply(int c)
 {
-	uint8_t *rwire = malloc(SOCKET_MTU_SZ);
-	knot_packet_t *reply = knot_packet_new();
-	if (!rwire || !reply) {
-		free(rwire);
-		knot_packet_free(&reply);
+	knot_pkt_t *pkt = knot_pkt_new(NULL, KNOT_WIRE_MAX_PKTSIZE, NULL);
+	if (!pkt) {
 		return KNOT_ENOMEM;
 	}
 
 	/* Read response packet. */
-	int n = tcp_recv(c, rwire, SOCKET_MTU_SZ, NULL);
+	int n = tcp_recv_msg(c, pkt->wire, pkt->max_size, NULL);
 	if (n <= 0) {
-		dbg_server("remote: couldn't receive response = %d\n", n);
-		knot_packet_free(&reply);
-		free(rwire);
+		dbg_server("remote: couldn't receive response = %s\n", knot_strerror(n));
+		knot_pkt_free(&pkt);
 		return KNOT_ECONN;
+	} else {
+		pkt->size = n;
 	}
 
 	/* Parse packet and check response. */
-	int ret = remote_parse(reply, rwire, n);
-	if (ret == KNOT_EOK) {
-		/* Check RCODE */
-		ret = knot_packet_rcode(reply);
-		switch(ret) {
-		case KNOT_RCODE_NOERROR:
-			if (knot_packet_authority_rrset_count(reply) > 0) {
-				ret = cmd_remote_print_reply(reply->authority[0]);
-			}
-			break;
-		case KNOT_RCODE_REFUSED:
-			ret = KNOT_EDENIED;
-			break;
-		default:
-			ret = KNOT_ERROR;
-			break;
-		}
+	int ret = remote_parse(pkt);
+	if (ret != KNOT_EOK) {
+		knot_pkt_free(&pkt);
+		return ret;
 	}
 
-	/* Response cleanup. */
-	knot_packet_free(&reply);
-	free(rwire);
+	/* Check RCODE */
+	const knot_pktsection_t *authority = knot_pkt_section(pkt, KNOT_AUTHORITY);
+	ret = knot_wire_get_rcode(pkt->wire);
+	switch(ret) {
+	case KNOT_RCODE_NOERROR:
+		if (authority->count > 0) {
+			ret = cmd_remote_print_reply(&authority->rr[0]);
+		}
+		break;
+	case KNOT_RCODE_REFUSED:
+		ret = KNOT_EDENIED;
+		break;
+	default:
+		ret = KNOT_ERROR;
+		break;
+	}
+
+	knot_pkt_free(&pkt);
 	return ret;
 }
 
@@ -186,82 +180,111 @@ static int cmd_remote(const char *cmd, uint16_t rrt, int argc, char *argv[])
 
 	/* Check remote address. */
 	conf_iface_t *r = conf()->ctl.iface;
-	if (!r || !r->address) {
-		log_server_error("No remote address for '%s' configured.\n",
-		                 cmd);
+	if (!r || r->addr.ss_family == AF_UNSPEC) {
+		log_error("no remote address for '%s' configured", cmd);
 		return 1;
 	}
 
 	/* Make query. */
-	uint8_t *buf = NULL;
-	size_t buflen = 0;
-	knot_packet_t *qr = remote_query(cmd, r->key);
-	if (!qr) {
-		log_server_warning("Could not prepare query for '%s'.\n",
-		                   cmd);
+	knot_pkt_t *pkt = remote_query(cmd, r->key);
+	if (!pkt) {
+		log_warning("failed to prepare query for '%s'", cmd);
 		return 1;
 	}
 
 	/* Build query data. */
-	knot_rrset_t *rr = NULL;
+	knot_pkt_begin(pkt, KNOT_AUTHORITY);
 	if (argc > 0) {
-		rr = remote_build_rr("data.", rrt);
+		knot_rrset_t rr;
+		int res = remote_build_rr(&rr, "data.", rrt);
+		if (res != KNOT_EOK) {
+			log_error("failed to create the query");
+			knot_pkt_free(&pkt);
+			return 1;
+		}
 		for (int i = 0; i < argc; ++i) {
 			switch(rrt) {
 			case KNOT_RRTYPE_NS:
-				remote_create_ns(rr, argv[i]);
+				remote_create_ns(&rr, argv[i]);
 				break;
 			case KNOT_RRTYPE_TXT:
 			default:
-				remote_create_txt(rr, argv[i], strlen(argv[i]));
+				remote_create_txt(&rr, argv[i], strlen(argv[i]));
 				break;
 			}
 		}
-		remote_query_append(qr, rr);
-	}
-
-	if (knot_packet_to_wire(qr, &buf, &buflen) != KNOT_EOK) {
-		knot_rrset_deep_free(&rr, 1);
-		knot_packet_free(&qr);
-		return 1;
+		res = knot_pkt_put(pkt, 0, &rr, KNOT_PF_FREE);
+		if (res != KNOT_EOK) {
+			log_error("failed to create the query");
+			knot_rrset_clear(&rr, NULL);
+			knot_pkt_free(&pkt);
+			return 1;
+		}
 	}
 
 	if (r->key) {
-		remote_query_sign(buf, &buflen, qr->max_size, r->key);
+		int res = remote_query_sign(pkt->wire, &pkt->size, pkt->max_size, r->key);
+		if (res != KNOT_EOK) {
+			log_error("failed to sign the packet");
+			knot_pkt_free(&pkt);
+			return 1;
+		}
 	}
 
-	/* Send query. */
-	int s = socket_create(r->family, SOCK_STREAM, 0);
-	int conn_state = socket_connect(s, r->family, r->address, r->port);
-	if (conn_state != KNOT_EOK || tcp_send(s, buf, buflen) <= 0) {
-		char portstr[32] = { '\0' };
-		if (r->family != AF_UNIX)
-			snprintf(portstr, sizeof(portstr), "@%d", r->port);
-		log_server_error("Couldn't connect to remote host "
-		                 "%s%s\n", r->address, portstr);
-		rc = 1;
+	dbg_server("%s: sending query size %zu\n", __func__, pkt->size);
+
+	/* Connect to remote. */
+	char addr_str[SOCKADDR_STRLEN] = {0};
+	sockaddr_tostr(&r->addr, addr_str, sizeof(addr_str));
+
+	int s = net_connected_socket(SOCK_STREAM, &r->addr, &r->via, 0);
+	if (s < 0) {
+		log_error("failed to connect to remote host '%s'", addr_str);
+		knot_pkt_free(&pkt);
+		return 1;
+	}
+
+	/* Wait for availability. */
+	struct pollfd pfd = { s, POLLOUT, 0 };
+	if (poll(&pfd, 1, conf()->max_conn_reply) != 1) {
+		log_error("failed to connect to remote host '%s'", addr_str);
+		close(s);
+		knot_pkt_free(&pkt);
+		return 1;
+	}
+
+	/* Send and free packet. */
+	int ret = tcp_send_msg(s, pkt->wire, pkt->size, NULL);
+	knot_pkt_free(&pkt);
+
+	/* Evaluate and wait for reply. */
+	if (ret <= 0) {
+		log_error("failed to connect to remote host '%s'", addr_str);
+		close(s);
+		return 1;
 	}
 
 	/* Wait for reply. */
-	if (rc == 0) {
-		int ret = KNOT_EOK;
-		while (ret == KNOT_EOK) {
-			ret = cmd_remote_reply(s);
-			if (ret != KNOT_EOK && ret != KNOT_ECONN) {
-				log_server_warning("Remote command reply: %s\n",
-				                   knot_strerror(ret));
+	ret = KNOT_EOK;
+	while (ret == KNOT_EOK) {
+		ret = cmd_remote_reply(s);
+		if (ret != KNOT_EOK) {
+			if (ret != KNOT_ECONN) {
+				log_notice("remote command reply: %s",
+				           knot_strerror(ret));
 				rc = 1;
 			}
+			break;
 		}
 	}
 
 	/* Cleanup. */
-	if (rc == 0) printf("\n");
-	knot_rrset_deep_free(&rr, 1);
+	if (rc == 0) {
+		printf("\n");
+	}
 
 	/* Close connection. */
-	socket_close(s);
-	knot_packet_free(&qr);
+	close(s);
 	return rc;
 }
 
@@ -284,7 +307,7 @@ static int tsig_parse_str(knot_tsig_key_t *key, const char *str)
 	if (s) {
 		*s++ = '\0';               /* Last part separator */
 		knot_lookup_table_t *alg = NULL;
-		alg = knot_lookup_by_name(knot_tsig_alg_dnames_str, h);
+		alg = knot_lookup_by_name(knot_tsig_alg_names, h);
 		if (alg) {
 			algorithm = alg->id;
 		} else {
@@ -348,7 +371,7 @@ static int tsig_parse_file(knot_tsig_key_t *k, const char *f)
 {
 	FILE* fp = fopen(f, "r");
 	if (!fp) {
-		log_server_error("Couldn't open key-file '%s'.\n", f);
+		log_error("failed to open key-file '%s'", f);
 		return KNOT_EINVAL;
 	}
 
@@ -368,9 +391,8 @@ static int tsig_parse_file(knot_tsig_key_t *k, const char *f)
 		}
 		if (c == '\n') {
 			if (k->name) {
-				log_server_error("Only 1 key definition "
-				                 "allowed in '%s'.\n",
-				                 f);
+				log_error("only one key definition allowed "
+				          "in '%s'", f);
 				ret = KNOT_EMALF;
 				break;
 			}
@@ -393,8 +415,7 @@ int main(int argc, char **argv)
 	/* Parse command line arguments */
 	int c = 0, li = 0, rc = 0;
 	unsigned flags = F_NULL;
-	char *config_fn = NULL;
-	char *default_config = conf_find_default();
+	const char *config_fn = conf_find_default();
 
 	/* Remote server descriptor. */
 	const char *r_addr = NULL;
@@ -411,7 +432,6 @@ int main(int argc, char **argv)
 		{"force", no_argument, 0, 'f'},
 		{"config", required_argument, 0, 'c'},
 		{"verbose", no_argument, 0, 'v'},
-		{"interactive", no_argument, 0, 'i'},
 		{"version", no_argument, 0, 'V'},
 		{"help", no_argument, 0, 'h'},
 		{"server", required_argument, 0, 's' },
@@ -421,7 +441,7 @@ int main(int argc, char **argv)
 		{0, 0, 0, 0}
 	};
 
-	while ((c = getopt_long(argc, argv, "s:p:y:k:fc:viVh", opts, &li)) != -1) {
+	while ((c = getopt_long(argc, argv, "s:p:y:k:fc:vVh", opts, &li)) != -1) {
 		switch (c) {
 		case 's':
 			r_addr = optarg;
@@ -432,16 +452,14 @@ int main(int argc, char **argv)
 		case 'y':
 			if (tsig_parse_str(&r_key, optarg) != KNOT_EOK) {
 				rc = 1;
-				log_server_error("Couldn't parse TSIG key '%s' "
-				                 "\n", optarg);
+				log_error("failed to parse TSIG key '%s'", optarg);
 				goto exit;
 			}
 			break;
 		case 'k':
 			if (tsig_parse_file(&r_key, optarg) != KNOT_EOK) {
 				rc = 1;
-				log_server_error("Couldn't parse TSIG key file "
-				                 "'%s'\n", optarg);
+				log_error("failed to parse TSIG key file '%s'", optarg);
 				goto exit;
 			}
 			break;
@@ -451,11 +469,8 @@ int main(int argc, char **argv)
 		case 'v':
 			flags |= F_VERBOSE;
 			break;
-		case 'i':
-			flags |= F_INTERACTIVE;
-			break;
 		case 'c':
-			config_fn = strdup(optarg);
+			config_fn = optarg;
 			break;
 		case 'V':
 			rc = 0;
@@ -491,31 +506,37 @@ int main(int argc, char **argv)
 
 	/* Command not found. */
 	if (!cmd->name) {
-		log_server_error("Invalid command: '%s'\n", argv[optind]);
+		log_error("invalid command: '%s'", argv[optind]);
 		rc = 1;
 		goto exit;
 	}
 
-	/* Open config, allow if not exists. */
+	/* Open config, create empty if not exists. */
 	if (conf_open(config_fn) != KNOT_EOK) {
-		if(conf_open(default_config) != KNOT_EOK) {
-			flags |= F_NOCONF;
-		}
+		s_config = conf_new("");
+		flags |= F_NOCONF;
+	}
+
+	/* Check if config file is required. */
+	if (has_flag(flags, F_NOCONF) && cmd->need_conf) {
+		log_error("failed to find a config file, refusing to continue");
+		rc = 1;
+		goto exit;
 	}
 
 	/* Create remote iface if not present in config. */
 	conf_iface_t *ctl_if = conf()->ctl.iface;
 	if (!ctl_if) {
 		ctl_if = malloc(sizeof(conf_iface_t));
-		assert(ctl_if);
-		conf()->ctl.iface = ctl_if;
 		memset(ctl_if, 0, sizeof(conf_iface_t));
+		conf()->ctl.iface = ctl_if;
 
 		/* Fill defaults. */
-		if (!r_addr)
+		if (!r_addr) {
 			r_addr = RUN_DIR "/knot.sock";
-		else if (r_port < 0)
+		} else if (r_port < 0) {
 			r_port = REMOTE_DPORT;
+		}
 	}
 
 	/* Install the key. */
@@ -525,25 +546,26 @@ int main(int argc, char **argv)
 
 	/* Override from command line. */
 	if (r_addr) {
-		free(ctl_if->address);
-		ctl_if->address = strdup(r_addr);
-		ctl_if->family = AF_INET;
-
 		/* Check for v6 address. */
-		if (strchr(r_addr, ':'))
-			ctl_if->family = AF_INET6;
+		int family = AF_INET;
+		if (strchr(r_addr, ':')) {
+			family = AF_INET6;
+		}
 
 		/* Is a valid UNIX socket or at least contains slash ? */
 		struct stat st;
 		bool has_slash = strchr(r_addr, '/') != NULL;
 		bool is_file = stat(r_addr, &st) == 0;
 		if (has_slash || (is_file && S_ISSOCK(st.st_mode))) {
-			ctl_if->family = AF_UNIX;
-			r_port = 0; /* Override. */
+			family = AF_UNIX;
 		}
 
+		sockaddr_set(&ctl_if->addr, family, r_addr, sockaddr_port(&ctl_if->addr));
+	}
+
+	if (r_port > 0) {
+		sockaddr_port_set(&ctl_if->addr, r_port);
 	}
-	if (r_port > -1) ctl_if->port = r_port;
 
 	/* Verbose mode. */
 	if (has_flag(flags, F_VERBOSE)) {
@@ -558,8 +580,6 @@ exit:
 	/* Finish */
 	knot_tsig_key_free(&r_key);
 	log_close();
-	free(config_fn);
-	free(default_config);
 	return rc;
 }
 
@@ -585,7 +605,7 @@ static int cmd_refresh(int argc, char *argv[], unsigned flags)
 {
 	UNUSED(flags);
 
-	if (has_flag(flags, F_FORCE)) {
+	if (flags & F_FORCE) {
 		return cmd_remote("retransfer", KNOT_RRTYPE_NS, argc, argv);
 	} else {
 		return cmd_remote("refresh", KNOT_RRTYPE_NS, argc, argv);
@@ -628,74 +648,75 @@ static int cmd_checkconf(int argc, char *argv[], unsigned flags)
 	UNUSED(argv);
 	UNUSED(flags);
 
-	/* Check config. */
-	if (has_flag(flags, F_NOCONF)) {
-		log_server_error("Couldn't parse config file, refusing to "
-		                 "continue.\n");
-		return 1;
-	} else {
-		log_server_info("OK, configuration is valid.\n");
-	}
+	log_info("configuration is valid");
 
 	return 0;
 }
 
+static bool fetch_zone(int argc, char *argv[], conf_zone_t *zone)
+{
+	/* Convert zone name to dname */
+	knot_dname_t *zone_name = knot_dname_from_str_alloc(zone->name);
+	if (zone_name == NULL) {
+		return false;
+	}
+	(void)knot_dname_to_lower(zone_name);
+
+	bool found = false;
+
+	int i = 0;
+	while (!found && i < argc) {
+		/* Convert the argument to dname */
+		knot_dname_t *arg_name = knot_dname_from_str_alloc(argv[i]);
+
+		if (arg_name != NULL) {
+			(void)knot_dname_to_lower(arg_name);
+			found = knot_dname_is_equal(zone_name, arg_name);
+		}
+
+		i++;
+		knot_dname_free(&arg_name, NULL);
+	}
+
+	knot_dname_free(&zone_name, NULL);
+	return found;
+}
+
 static int cmd_checkzone(int argc, char *argv[], unsigned flags)
 {
 	UNUSED(flags);
 
 	/* Zone checking */
 	int rc = 0;
-	node_t *n = 0;
 
 	/* Generate databases for all zones */
-	WALK_LIST(n, conf()->zones) {
-		/* Fetch zone */
-		conf_zone_t *zone = (conf_zone_t *) n;
-		int zone_match = 0;
-		for (unsigned i = 0; i < (unsigned)argc; ++i) {
-			size_t len = strlen(zone->name);
+	const bool sorted = false;
+	hattrie_iter_t *z_iter = hattrie_iter_begin(conf()->zones, sorted);
+	if (z_iter == NULL) {
+		return KNOT_ERROR;
+	}
+	for (; !hattrie_iter_finished(z_iter); hattrie_iter_next(z_iter)) {
+		conf_zone_t *zone = (conf_zone_t *)*hattrie_iter_val(z_iter);
 
-			/* All (except root) without final dot */
-			if (len > 1) {
-				len -= 1;
-			}
-			if (strncmp(zone->name, argv[i], len) == 0) {
-				zone_match = 1;
-				break;
-			}
-		}
+		/* Fetch zone */
+		int zone_match = fetch_zone(argc, argv, zone);
 
 		if (!zone_match && argc > 0) {
-			/* WALK_LIST is a for-cycle. */
+			conf_free_zone(zone);
 			continue;
 		}
 
 		/* Create zone loader context. */
-		zloader_t *l = NULL;
-		int ret = knot_zload_open(&l, zone->file, zone->name,
-		                          zone->enable_checks);
-		if (ret != KNOT_EOK) {
-			log_zone_error("Could not open zone %s (%s).\n",
-			               zone->name, knot_strerror(ret));
-			knot_zload_close(l);
+		zone_contents_t *loaded_zone = zone_load_contents(zone);
+		if (loaded_zone == NULL) {
 			rc = 1;
 			continue;
 		}
 
-		knot_zone_t *z = knot_zload_load(l);
-		if (z == NULL) {
-			log_zone_error("Loading of zone %s failed.\n",
-			               zone->name);
-			knot_zload_close(l);
-			rc = 1;
-			continue;
-		}
-
-		knot_zone_deep_free(&z);
-		knot_zload_close(l);
-		log_zone_info("Zone %s OK.\n", zone->name);
+		log_zone_str_info(zone->name, "zone is valid");
+		zone_contents_deep_free(&loaded_zone);
 	}
+	hattrie_iter_free(z_iter);
 
 	return rc;
 }
@@ -705,30 +726,22 @@ static int cmd_memstats(int argc, char *argv[], unsigned flags)
 	UNUSED(flags);
 
 	/* Zone checking */
-	int rc = 0;
-	node_t *n = 0;
-	size_t total_size = 0;
+	double total_size = 0;
 
 	/* Generate databases for all zones */
-	WALK_LIST(n, conf()->zones) {
-		/* Fetch zone */
-		conf_zone_t *zone = (conf_zone_t *) n;
-		int zone_match = 0;
-		for (unsigned i = 0; i < (unsigned)argc; ++i) {
-			size_t len = strlen(zone->name);
+	const bool sorted = false;
+	hattrie_iter_t *z_iter = hattrie_iter_begin(conf()->zones, sorted);
+	if (z_iter == NULL) {
+		return KNOT_ERROR;
+	}
+	for (; !hattrie_iter_finished(z_iter); hattrie_iter_next(z_iter)) {
+		conf_zone_t *zone = (conf_zone_t *)*hattrie_iter_val(z_iter);
 
-			/* All (except root) without final dot */
-			if (len > 1) {
-				len -= 1;
-			}
-			if (strncmp(zone->name, argv[i], len) == 0) {
-				zone_match = 1;
-				break;
-			}
-		}
+		/* Fetch zone */
+		int zone_match = fetch_zone(argc, argv, zone);
 
 		if (!zone_match && argc > 0) {
-			/* WALK_LIST is a for-cycle. */
+			conf_free_zone(zone);
 			continue;
 		}
 
@@ -740,63 +753,62 @@ static int cmd_memstats(int argc, char *argv[], unsigned flags)
 
 		/* Init memory estimation context. */
 		zone_estim_t est = {.node_table = hattrie_create_n(TRIE_BUCKET_SIZE, &mem_ctx),
-		                    .dname_size = 0, .rrset_size = 0,
-		                    .node_size = 0, .ahtable_size = 0,
-		                    .rdata_size = 0, .record_count = 0 };
+		                    .dname_size = 0, .node_size = 0,
+		                    .htable_size = 0, .rdata_size = 0,
+		                    .record_count = 0 };
 		if (est.node_table == NULL) {
-			log_server_error("Not enough memory.\n");
+			log_error("not enough memory");
+			break;
+		}
+
+		/* Create zone scanner. */
+		zs_scanner_t *zs = zs_scanner_create(zone->name,
+		                                     KNOT_CLASS_IN, 3600,
+		                                     estimator_rrset_memsize_wrap,
+		                                     process_error,
+		                                     &est);
+		if (zs == NULL) {
+			log_zone_str_error(zone->name, "failed to load zone");
+			hattrie_free(est.node_table);
 			continue;
 		}
 
-		/* Create file loader. */
-		file_loader_t *loader = file_loader_create(zone->file, zone->name,
-		                                           KNOT_CLASS_IN, 3600,
-		                                           estimator_rrset_memsize_wrap,
-		                                           process_error,
-		                                           &est);
-		if (loader == NULL) {
-			rc = 1;
-			log_zone_error("Could not load zone %s\n", zone->name);
-			hattrie_apply_rev(est.node_table, estimator_free_trie_node, NULL);
-			hattrie_free(est.node_table);
-			break;
-		}
-
 		/* Do a parser run, but do not actually create the zone. */
-		int ret = file_loader_process(loader);
-		if (ret != KNOT_EOK) {
-			rc = 1;
-			log_zone_error("Failed to parse zone %s.\n", zone->name);
+		int ret = zs_scanner_parse_file(zs, zone->file);
+		if (ret != 0) {
+			log_zone_str_error(zone->name, "failed to parse zone");
 			hattrie_apply_rev(est.node_table, estimator_free_trie_node, NULL);
 			hattrie_free(est.node_table);
-			file_loader_free(loader);
-			break;
+			zs_scanner_free(zs);
+			continue;
 		}
 
 		/* Only size of ahtables inside trie's nodes is missing. */
-		assert(est.ahtable_size == 0);
-		est.ahtable_size = estimator_trie_ahtable_memsize(est.node_table);
+		assert(est.htable_size == 0);
+		est.htable_size = estimator_trie_htable_memsize(est.node_table);
 
 		/* Cleanup */
 		hattrie_apply_rev(est.node_table, estimator_free_trie_node, NULL);
 		hattrie_free(est.node_table);
 
-		size_t zone_size = (size_t)(((double)(est.rdata_size +
+		double zone_size = ((double)(est.rdata_size +
 		                   est.node_size +
-		                   est.rrset_size +
 		                   est.dname_size +
-		                   est.ahtable_size +
-		                   malloc_size) * ESTIMATE_MAGIC) / (1024.0 * 1024.0));
+		                   est.htable_size +
+		                   malloc_size) * ESTIMATE_MAGIC) / (1024.0 * 1024.0);
 
-		log_zone_info("Zone %s: %zu RRs, used memory estimation is %zuMB.\n",
-		              zone->name, est.record_count, zone_size);
-		file_loader_free(loader);
+		log_zone_str_info(zone->name, "%zu RRs, used memory estimation is %zu MB",
+		                  est.record_count, (size_t)zone_size);
+		zs_scanner_free(zs);
 		total_size += zone_size;
+		conf_free_zone(zone);
+	}
+	hattrie_iter_free(z_iter);
+
+	if (argc == 0) { // for all zones
+		log_info("estimated memory consumption for all zones is %zu MB",
+		         (size_t)total_size);
 	}
 
-	if (rc == 0 && argc == 0) { // for all zones
-		log_zone_info("Estimated memory consumption for all zones is %zuMB.\n", total_size);
-	}
-
-	return rc;
+	return 0;
 }
