@@ -14,7 +14,6 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#include <config.h>
 #include "libknot/rrset-dump.h"
 
 #include <stdlib.h>			// free
@@ -29,10 +28,10 @@
 #include <netinet/in.h>			// in_addr (BSD)
 #include <arpa/inet.h>			// ntohs
 
-#include "common/errcode.h"		// KNOT_EOK
+#include "libknot/errcode.h"		// KNOT_EOK
 #include "common/base64.h"		// base64
 #include "common/base32hex.h"		// base32hex
-#include "common/descriptor.h"		// KNOT_RRTYPE
+#include "libknot/descriptor.h"		// KNOT_RRTYPE
 #include "libknot/dnssec/key.h"		// knot_keytag
 #include "libknot/consts.h"		// knot_rcode_names
 #include "libknot/util/utils.h"		// knot_wire_read_u16
@@ -58,6 +57,7 @@ const knot_dump_style_t KNOT_DUMP_STYLE_DEFAULT = {
 	.show_class = false,
 	.show_ttl = true,
 	.verbose = false,
+	.empty_ttl = false,
 	.human_ttl = false,
 	.human_tmstamp = true,
 	.ascii_to_idn = NULL
@@ -67,8 +67,6 @@ static void dump_string(rrset_dump_params_t *p, const char *str)
 {
 	size_t in_len = strlen(str);
 
-	p->ret = -1;
-
 	// Check input size (+ 1 termination).
 	if (in_len >= p->out_max) {
 		return;
@@ -92,8 +90,6 @@ static void wire_num8_to_str(rrset_dump_params_t *p)
 	size_t  in_len = sizeof(data);
 	size_t  out_len = 0;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -121,8 +117,6 @@ static void wire_num16_to_str(rrset_dump_params_t *p)
 	size_t   in_len = sizeof(data);
 	size_t   out_len = 0;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -153,8 +147,6 @@ static void wire_num32_to_str(rrset_dump_params_t *p)
 	size_t   in_len = sizeof(data);
 	size_t   out_len = 0;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -185,8 +177,6 @@ static void wire_num48_to_str(rrset_dump_params_t *p)
 	size_t   in_len = 6;
 	size_t   out_len = 0;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -217,8 +207,6 @@ static void wire_ipv4_to_str(rrset_dump_params_t *p)
 	size_t in_len = sizeof(addr4.s_addr);
 	size_t out_len = 0;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -250,8 +238,6 @@ static void wire_ipv6_to_str(rrset_dump_params_t *p)
 	size_t in_len = sizeof(addr6.s6_addr);
 	size_t out_len = 0;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -284,8 +270,6 @@ static void wire_type_to_str(rrset_dump_params_t *p)
 	size_t   in_len = sizeof(data);
 	size_t   out_len = 0;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -317,7 +301,6 @@ static void wire_type_to_str(rrset_dump_params_t *p)
 	p->ret = 0;
 }
 
-
 static int hex_encode(const uint8_t  *in,
                       const uint32_t in_len,
                       uint8_t        *out,
@@ -366,8 +349,6 @@ static void wire_data_encode_to_str(rrset_dump_params_t *p,
 	int    ret;
 	size_t in_len = p->in_max;
 
-	p->ret = -1;
-
 	// One-line vs multi-line mode.
 	if (p->style->wrap == false) {
 		// Encode data directly to the output.
@@ -381,11 +362,11 @@ static void wire_data_encode_to_str(rrset_dump_params_t *p,
 		p->out_max -= out_len;
 		p->total += out_len;
 	} else {
-		int  src_begin, src_len;
-		char *buf;
+		int     src_begin;
+		uint8_t *buf;
 
 		// Encode data to the temporary buffer.
-		ret = enc_alloc(p->in, in_len, (uint8_t **)&buf);
+		ret = enc_alloc(p->in, in_len, &buf);
 		if (ret <= 0) {
 			return;
 		}
@@ -402,8 +383,8 @@ static void wire_data_encode_to_str(rrset_dump_params_t *p,
 			}
 
 			// Compute block length (the last one can be shorter).
-			src_len = (ret - src_begin) < BLOCK_WIDTH ?
-			          (ret - src_begin) : BLOCK_WIDTH;
+			int src_len = (ret - src_begin) < BLOCK_WIDTH ?
+			              (ret - src_begin) : BLOCK_WIDTH;
 
 			if ((size_t)src_len > p->out_max) {
 				free(buf);
@@ -443,8 +424,6 @@ static void wire_len_data_encode_to_str(rrset_dump_params_t *p,
 {
 	size_t in_len;
 
-	p->ret = -1;
-
 	// First len_len bytes are data length.
 	if (p->in_max < len_len) {
 		return;
@@ -527,8 +506,6 @@ static void wire_len_data_encode_to_str(rrset_dump_params_t *p,
 
 static void wire_text_to_str(rrset_dump_params_t *p)
 {
-	p->ret = -1;
-
 	// First byte is string length.
 	if (p->in_max < 1) {
 		return;
@@ -609,8 +586,6 @@ static void wire_timestamp_to_str(rrset_dump_params_t *p)
 	size_t   out_len = 0;
 	int      ret;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -718,8 +693,6 @@ static void wire_ttl_to_str(rrset_dump_params_t *p)
 	size_t   out_len = 0;
 	int      ret;
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -762,8 +735,6 @@ static void wire_bitmap_to_str(rrset_dump_params_t *p)
 	size_t in_len = p->in_max;
 	size_t out_len = 0;
 
-	p->ret = -1;
-
 	// Loop over bitmap window array (can be empty).
 	while (i < in_len) {
 		// First byte is window number.
@@ -821,8 +792,6 @@ static void wire_bitmap_to_str(rrset_dump_params_t *p)
 
 static void wire_dname_to_str(rrset_dump_params_t *p)
 {
-	p->ret = -1;
-
 	int in_len = knot_dname_size(p->in);
 	if (in_len < 0) {
 		return;
@@ -835,16 +804,23 @@ static void wire_dname_to_str(rrset_dump_params_t *p)
 	}
 
 	// Write dname string.
-	char *dname_str = knot_dname_to_str(p->in);
-	if (p->style->ascii_to_idn != NULL) {
+	if (p->style->ascii_to_idn == NULL) {
+		char *dname_str = knot_dname_to_str(p->out, p->in, p->out_max);
+		if (dname_str == NULL) {
+			return;
+		}
+		out_len = strlen(dname_str);
+	} else {
+		char *dname_str = knot_dname_to_str_alloc(p->in);
 		p->style->ascii_to_idn(&dname_str);
+
+		int ret = snprintf(p->out, p->out_max, "%s", dname_str);
+		free(dname_str);
+		if (ret < 0 || (size_t)ret >= p->out_max) {
+			return;
+		}
+		out_len = ret;
 	}
-	int ret = snprintf(p->out, p->out_max, "%s", dname_str);
-	free(dname_str);
-	if (ret < 0 || (size_t)ret >= p->out_max) {
-		return;
-	}
-	out_len = ret;
 
 	// Fill in output.
 	p->in += in_len;
@@ -862,8 +838,6 @@ static void wire_apl_to_str(rrset_dump_params_t *p)
 	int    ret;
 	size_t out_len = 0;
 
-	p->ret = -1;
-
 	// Input check: family(2B) + prefix(1B) + afdlen(1B).
 	if (p->in_max < 4) {
 		return;
@@ -958,8 +932,6 @@ static void wire_loc_to_str(rrset_dump_params_t *p)
 	int    ret;
 	size_t in_len = 16;
 
-	p->ret = -1;
-
 	// Check input size (1 LOC = 16 B).
 	if (in_len > p->in_max) {
 		return;
@@ -1069,8 +1041,6 @@ static void wire_loc_to_str(rrset_dump_params_t *p)
 
 static void wire_gateway_to_str(rrset_dump_params_t *p)
 {
-	p->ret = -1;
-
 	// Input check: type(1B) + algo(1B).
 	if (p->in_max < 2) {
 		return;
@@ -1147,8 +1117,6 @@ static void wire_gateway_to_str(rrset_dump_params_t *p)
 
 static void wire_l64_to_str(rrset_dump_params_t *p)
 {
-	p->ret = -1;
-
 	// Check input size (64-bit identifier).
 	if (p->in_max != 8) {
 		return;
@@ -1180,8 +1148,6 @@ static void wire_l64_to_str(rrset_dump_params_t *p)
 
 static void wire_eui_to_str(rrset_dump_params_t *p)
 {
-	p->ret = -1;
-
 	// Data can't have zero length.
 	if (p->in_max < 2) {
 		return;
@@ -1217,8 +1183,6 @@ static void wire_rcode_to_str(rrset_dump_params_t *p)
 	size_t   in_len = sizeof(data);
 	const char *rcode_str = "NULL";
 
-	p->ret = -1;
-
 	// Check input size.
 	if (in_len > p->in_max) {
 		return;
@@ -1251,8 +1215,6 @@ static void wire_unknown_to_str(rrset_dump_params_t *p)
 	size_t in_len = p->in_max;
 	size_t out_len = 0;
 
-	p->ret = -1;
-
 	// Write unknown length header.
 	if (in_len > 0) {
 		ret = snprintf(p->out, p->out_max, "\\# %zu ", in_len);
@@ -1308,60 +1270,55 @@ static void dnskey_info(const uint8_t *rdata,
 	}
 }
 
-#define DUMP_PARAMS	const uint8_t *in, const size_t in_len, char *out, \
-			const size_t out_max, const knot_dump_style_t *style
-#define DUMP_INIT	rrset_dump_params_t p = { .style = style, .in = in, \
-			.in_max = in_len, .out = out, .out_max = out_max };
-#define	DUMP_END	return (p.in_max == 0 ? (int)p.total : KNOT_EPARSEFAIL);
+#define DUMP_PARAMS	rrset_dump_params_t *const p
+#define	DUMP_END	return (p->in_max == 0 ? (int)p->total : KNOT_EPARSEFAIL);
 
-#define CHECK_RET(p)	if (p.ret != 0) return -1;
+#define CHECK_RET(p)	if (p->ret != 0) return -1;
 
-#define WRAP_INIT	dump_string(&p, "(" BLOCK_INDENT); CHECK_RET(p);
-#define WRAP_END	dump_string(&p, BLOCK_INDENT ")"); CHECK_RET(p);
-#define WRAP_LINE	dump_string(&p, BLOCK_INDENT); CHECK_RET(p);
+#define WRAP_INIT	dump_string(p, "(" BLOCK_INDENT); CHECK_RET(p);
+#define WRAP_END	dump_string(p, BLOCK_INDENT ")"); CHECK_RET(p);
+#define WRAP_LINE	dump_string(p, BLOCK_INDENT); CHECK_RET(p);
 
-#define COMMENT(s)	if (p.style->verbose) { \
-			    dump_string(&p, " ; "); CHECK_RET(p); \
-			    dump_string(&p, s); CHECK_RET(p); \
+#define COMMENT(s)	if (p->style->verbose) { \
+			    dump_string(p, " ; "); CHECK_RET(p); \
+			    dump_string(p, s); CHECK_RET(p); \
 			}
 
-#define DUMP_SPACE	dump_string(&p, " "); CHECK_RET(p);
-#define DUMP_NUM8	wire_num8_to_str(&p); CHECK_RET(p);
-#define DUMP_NUM16	wire_num16_to_str(&p); CHECK_RET(p);
-#define DUMP_NUM32	wire_num32_to_str(&p); CHECK_RET(p);
-#define DUMP_NUM48	wire_num48_to_str(&p); CHECK_RET(p);
-#define DUMP_DNAME	wire_dname_to_str(&p); CHECK_RET(p);
-#define DUMP_TIME	wire_ttl_to_str(&p); CHECK_RET(p);
-#define DUMP_TIMESTAMP	wire_timestamp_to_str(&p); CHECK_RET(p);
-#define DUMP_IPV4	wire_ipv4_to_str(&p); CHECK_RET(p);
-#define DUMP_IPV6	wire_ipv6_to_str(&p); CHECK_RET(p);
-#define DUMP_TYPE	wire_type_to_str(&p); CHECK_RET(p);
-#define DUMP_HEX	wire_data_encode_to_str(&p, &hex_encode, \
+#define DUMP_SPACE	dump_string(p, " "); CHECK_RET(p);
+#define DUMP_NUM8	wire_num8_to_str(p); CHECK_RET(p);
+#define DUMP_NUM16	wire_num16_to_str(p); CHECK_RET(p);
+#define DUMP_NUM32	wire_num32_to_str(p); CHECK_RET(p);
+#define DUMP_NUM48	wire_num48_to_str(p); CHECK_RET(p);
+#define DUMP_DNAME	wire_dname_to_str(p); CHECK_RET(p);
+#define DUMP_TIME	wire_ttl_to_str(p); CHECK_RET(p);
+#define DUMP_TIMESTAMP	wire_timestamp_to_str(p); CHECK_RET(p);
+#define DUMP_IPV4	wire_ipv4_to_str(p); CHECK_RET(p);
+#define DUMP_IPV6	wire_ipv6_to_str(p); CHECK_RET(p);
+#define DUMP_TYPE	wire_type_to_str(p); CHECK_RET(p);
+#define DUMP_HEX	wire_data_encode_to_str(p, &hex_encode, \
 				&hex_encode_alloc); CHECK_RET(p);
-#define DUMP_BASE64	wire_data_encode_to_str(&p, &base64_encode, \
+#define DUMP_BASE64	wire_data_encode_to_str(p, &base64_encode, \
 				&base64_encode_alloc); CHECK_RET(p);
-#define DUMP_HASH	wire_len_data_encode_to_str(&p, &base32hex_encode, \
+#define DUMP_HASH	wire_len_data_encode_to_str(p, &base32hex_encode, \
 				1, false, ""); CHECK_RET(p);
-#define DUMP_SALT	wire_len_data_encode_to_str(&p, &hex_encode, \
+#define DUMP_SALT	wire_len_data_encode_to_str(p, &hex_encode, \
 				1, false, "-"); CHECK_RET(p);
-#define DUMP_TSIG_DGST	wire_len_data_encode_to_str(&p, &base64_encode, \
+#define DUMP_TSIG_DGST	wire_len_data_encode_to_str(p, &base64_encode, \
 				2, true, ""); CHECK_RET(p);
-#define DUMP_TSIG_DATA	wire_len_data_encode_to_str(&p, &hex_encode, \
+#define DUMP_TSIG_DATA	wire_len_data_encode_to_str(p, &hex_encode, \
 				2, true, ""); CHECK_RET(p);
-#define DUMP_TEXT	wire_text_to_str(&p); CHECK_RET(p);
-#define DUMP_BITMAP	wire_bitmap_to_str(&p); CHECK_RET(p);
-#define DUMP_APL	wire_apl_to_str(&p); CHECK_RET(p);
-#define DUMP_LOC	wire_loc_to_str(&p); CHECK_RET(p);
-#define DUMP_GATEWAY	wire_gateway_to_str(&p); CHECK_RET(p);
-#define DUMP_L64	wire_l64_to_str(&p); CHECK_RET(p);
-#define DUMP_EUI	wire_eui_to_str(&p); CHECK_RET(p);
-#define DUMP_RCODE	wire_rcode_to_str(&p); CHECK_RET(p);
-#define DUMP_UNKNOWN	wire_unknown_to_str(&p); CHECK_RET(p);
+#define DUMP_TEXT	wire_text_to_str(p); CHECK_RET(p);
+#define DUMP_BITMAP	wire_bitmap_to_str(p); CHECK_RET(p);
+#define DUMP_APL	wire_apl_to_str(p); CHECK_RET(p);
+#define DUMP_LOC	wire_loc_to_str(p); CHECK_RET(p);
+#define DUMP_GATEWAY	wire_gateway_to_str(p); CHECK_RET(p);
+#define DUMP_L64	wire_l64_to_str(p); CHECK_RET(p);
+#define DUMP_EUI	wire_eui_to_str(p); CHECK_RET(p);
+#define DUMP_RCODE	wire_rcode_to_str(p); CHECK_RET(p);
+#define DUMP_UNKNOWN	wire_unknown_to_str(p); CHECK_RET(p);
 
 static int dump_a(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_IPV4;
 
 	DUMP_END;
@@ -1369,8 +1326,6 @@ static int dump_a(DUMP_PARAMS)
 
 static int dump_ns(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_DNAME;
 
 	DUMP_END;
@@ -1378,9 +1333,7 @@ static int dump_ns(DUMP_PARAMS)
 
 static int dump_soa(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_DNAME; DUMP_SPACE;
 		DUMP_DNAME; DUMP_SPACE; WRAP_INIT;
 		DUMP_NUM32; COMMENT("serial"); WRAP_LINE;
@@ -1403,8 +1356,6 @@ static int dump_soa(DUMP_PARAMS)
 
 static int dump_hinfo(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_TEXT; DUMP_SPACE;
 	DUMP_TEXT;
 
@@ -1413,8 +1364,6 @@ static int dump_hinfo(DUMP_PARAMS)
 
 static int dump_minfo(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_DNAME; DUMP_SPACE;
 	DUMP_DNAME;
 
@@ -1423,8 +1372,6 @@ static int dump_minfo(DUMP_PARAMS)
 
 static int dump_mx(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_NUM16; DUMP_SPACE;
 	DUMP_DNAME;
 
@@ -1433,13 +1380,11 @@ static int dump_mx(DUMP_PARAMS)
 
 static int dump_txt(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	// First text string.
 	DUMP_TEXT;
 
 	// Other text strings if any.
-	while (p.in_max > 0) {
+	while (p->in_max > 0) {
 		DUMP_SPACE; DUMP_TEXT;
 	}
 
@@ -1448,11 +1393,9 @@ static int dump_txt(DUMP_PARAMS)
 
 static int dump_dnskey(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		char info[512] = "";
-		dnskey_info(in, in_len, info, sizeof(info));
+		dnskey_info(p->in, p->in_max, info, sizeof(info));
 
 		DUMP_NUM16; DUMP_SPACE;
 		DUMP_NUM8;  DUMP_SPACE;
@@ -1471,8 +1414,6 @@ static int dump_dnskey(DUMP_PARAMS)
 
 static int dump_aaaa(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_IPV6;
 
 	DUMP_END;
@@ -1480,8 +1421,6 @@ static int dump_aaaa(DUMP_PARAMS)
 
 static int dump_loc(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_LOC;
 
 	DUMP_END;
@@ -1489,8 +1428,6 @@ static int dump_loc(DUMP_PARAMS)
 
 static int dump_srv(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_NUM16; DUMP_SPACE;
 	DUMP_NUM16; DUMP_SPACE;
 	DUMP_NUM16; DUMP_SPACE;
@@ -1501,8 +1438,6 @@ static int dump_srv(DUMP_PARAMS)
 
 static int dump_naptr(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_NUM16; DUMP_SPACE;
 	DUMP_NUM16; DUMP_SPACE;
 	DUMP_TEXT;  DUMP_SPACE;
@@ -1515,9 +1450,7 @@ static int dump_naptr(DUMP_PARAMS)
 
 static int dump_cert(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_NUM16;  DUMP_SPACE;
 		DUMP_NUM16;  DUMP_SPACE;
 		DUMP_NUM8;   DUMP_SPACE; WRAP_INIT;
@@ -1535,11 +1468,9 @@ static int dump_cert(DUMP_PARAMS)
 
 static int dump_apl(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	// Print list of APLs (empty list is allowed).
-	while (p.in_max > 0) {
-		if (p.total > 0) {
+	while (p->in_max > 0) {
+		if (p->total > 0) {
 			DUMP_SPACE;
 		}
 		DUMP_APL;
@@ -1550,9 +1481,7 @@ static int dump_apl(DUMP_PARAMS)
 
 static int dump_ds(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_NUM16; DUMP_SPACE;
 		DUMP_NUM8;  DUMP_SPACE;
 		DUMP_NUM8;  DUMP_SPACE; WRAP_INIT;
@@ -1570,9 +1499,7 @@ static int dump_ds(DUMP_PARAMS)
 
 static int dump_sshfp(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_NUM8; DUMP_SPACE;
 		DUMP_NUM8; DUMP_SPACE; WRAP_INIT;
 		DUMP_HEX;
@@ -1588,9 +1515,7 @@ static int dump_sshfp(DUMP_PARAMS)
 
 static int dump_ipseckey(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_NUM8; DUMP_SPACE; WRAP_INIT;
 		DUMP_GATEWAY;
 		WRAP_END;
@@ -1604,9 +1529,7 @@ static int dump_ipseckey(DUMP_PARAMS)
 
 static int dump_rrsig(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_TYPE;   DUMP_SPACE;
 		DUMP_NUM8;   DUMP_SPACE;
 		DUMP_NUM8;   DUMP_SPACE;
@@ -1634,8 +1557,6 @@ static int dump_rrsig(DUMP_PARAMS)
 
 static int dump_nsec(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_DNAME; DUMP_SPACE;
 	DUMP_BITMAP;
 
@@ -1644,9 +1565,7 @@ static int dump_nsec(DUMP_PARAMS)
 
 static int dump_dhcid(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		WRAP_INIT;
 		DUMP_BASE64;
 		WRAP_END;
@@ -1659,9 +1578,7 @@ static int dump_dhcid(DUMP_PARAMS)
 
 static int dump_nsec3(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_NUM8;   DUMP_SPACE;
 		DUMP_NUM8;   DUMP_SPACE;
 		DUMP_NUM16;  DUMP_SPACE;
@@ -1683,8 +1600,6 @@ static int dump_nsec3(DUMP_PARAMS)
 
 static int dump_nsec3param(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_NUM8;  DUMP_SPACE;
 	DUMP_NUM8;  DUMP_SPACE;
 	DUMP_NUM16; DUMP_SPACE;
@@ -1695,9 +1610,7 @@ static int dump_nsec3param(DUMP_PARAMS)
 
 static int dump_tlsa(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_NUM8; DUMP_SPACE;
 		DUMP_NUM8; DUMP_SPACE;
 		DUMP_NUM8; DUMP_SPACE; WRAP_INIT;
@@ -1715,8 +1628,6 @@ static int dump_tlsa(DUMP_PARAMS)
 
 static int dump_l64(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_NUM16; DUMP_SPACE;
 	DUMP_L64;
 
@@ -1725,8 +1636,6 @@ static int dump_l64(DUMP_PARAMS)
 
 static int dump_l32(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_NUM16; DUMP_SPACE;
 	DUMP_IPV4;
 
@@ -1735,8 +1644,6 @@ static int dump_l32(DUMP_PARAMS)
 
 static int dump_eui(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
 	DUMP_EUI;
 
 	DUMP_END;
@@ -1744,9 +1651,7 @@ static int dump_eui(DUMP_PARAMS)
 
 static int dump_tsig(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		DUMP_DNAME; DUMP_SPACE;
 		DUMP_NUM48; DUMP_SPACE;
 		DUMP_NUM16; DUMP_SPACE; WRAP_INIT;
@@ -1770,9 +1675,7 @@ static int dump_tsig(DUMP_PARAMS)
 
 static int dump_unknown(DUMP_PARAMS)
 {
-	DUMP_INIT;
-
-	if (p.style->wrap) {
+	if (p->style->wrap) {
 		WRAP_INIT;
 		DUMP_UNKNOWN;
 		WRAP_END;
@@ -1793,107 +1696,124 @@ int knot_rrset_txt_dump_data(const knot_rrset_t      *rrset,
 		return KNOT_EINVAL;
 	}
 
-	uint8_t *data = knot_rrset_get_rdata(rrset, pos);
-	size_t  data_len = rrset_rdata_item_size(rrset, pos);
+	const knot_rdata_t *rr_data = knot_rdataset_at(&rrset->rrs, pos);
+	uint8_t *data = knot_rdata_data(rr_data);
+	uint16_t data_len = knot_rdata_rdlen(rr_data);
+
+	rrset_dump_params_t p = {
+		.style = style,
+		.in = data,
+		.in_max = data_len,
+		.out = dst,
+		.out_max = maxlen,
+		.total = 0,
+		.ret = -1
+	};
 
 	int ret = 0;
 
-	switch (knot_rrset_type(rrset)) {
+	if (data_len == 0 && rrset->rclass != KNOT_CLASS_IN) {
+		return ret;
+	}
+
+	switch (rrset->type) {
 		case KNOT_RRTYPE_A:
-			ret = dump_a(data, data_len, dst, maxlen, style);
+			ret = dump_a(&p);
 			break;
 		case KNOT_RRTYPE_NS:
 		case KNOT_RRTYPE_CNAME:
 		case KNOT_RRTYPE_PTR:
 		case KNOT_RRTYPE_DNAME:
-			ret = dump_ns(data, data_len, dst, maxlen, style);
+			ret = dump_ns(&p);
 			break;
 		case KNOT_RRTYPE_SOA:
-			ret = dump_soa(data, data_len, dst, maxlen, style);
+			ret = dump_soa(&p);
 			break;
 		case KNOT_RRTYPE_HINFO:
-			ret = dump_hinfo(data, data_len, dst, maxlen, style);
+			ret = dump_hinfo(&p);
 			break;
 		case KNOT_RRTYPE_MINFO:
 		case KNOT_RRTYPE_RP:
-			ret = dump_minfo(data, data_len, dst, maxlen, style);
+			ret = dump_minfo(&p);
 			break;
 		case KNOT_RRTYPE_MX:
 		case KNOT_RRTYPE_AFSDB:
 		case KNOT_RRTYPE_RT:
 		case KNOT_RRTYPE_KX:
 		case KNOT_RRTYPE_LP:
-			ret = dump_mx(data, data_len, dst, maxlen, style);
+			ret = dump_mx(&p);
 			break;
 		case KNOT_RRTYPE_TXT:
 		case KNOT_RRTYPE_SPF:
-			ret = dump_txt(data, data_len, dst, maxlen, style);
+			ret = dump_txt(&p);
 			break;
 		case KNOT_RRTYPE_KEY:
 		case KNOT_RRTYPE_DNSKEY:
-			ret = dump_dnskey(data, data_len, dst, maxlen, style);
+		case KNOT_RRTYPE_CDNSKEY:
+			ret = dump_dnskey(&p);
 			break;
 		case KNOT_RRTYPE_AAAA:
-			ret = dump_aaaa(data, data_len, dst, maxlen, style);
+			ret = dump_aaaa(&p);
 			break;
 		case KNOT_RRTYPE_LOC:
-			ret = dump_loc(data, data_len, dst, maxlen, style);
+			ret = dump_loc(&p);
 			break;
 		case KNOT_RRTYPE_SRV:
-			ret = dump_srv(data, data_len, dst, maxlen, style);
+			ret = dump_srv(&p);
 			break;
 		case KNOT_RRTYPE_NAPTR:
-			ret = dump_naptr(data, data_len, dst, maxlen, style);
+			ret = dump_naptr(&p);
 			break;
 		case KNOT_RRTYPE_CERT:
-			ret = dump_cert(data, data_len, dst, maxlen, style);
+			ret = dump_cert(&p);
 			break;
 		case KNOT_RRTYPE_APL:
-			ret = dump_apl(data, data_len, dst, maxlen, style);
+			ret = dump_apl(&p);
 			break;
 		case KNOT_RRTYPE_DS:
-			ret = dump_ds(data, data_len, dst, maxlen, style);
+		case KNOT_RRTYPE_CDS:
+			ret = dump_ds(&p);
 			break;
 		case KNOT_RRTYPE_SSHFP:
-			ret = dump_sshfp(data, data_len, dst, maxlen, style);
+			ret = dump_sshfp(&p);
 			break;
 		case KNOT_RRTYPE_IPSECKEY:
-			ret = dump_ipseckey(data, data_len, dst, maxlen, style);
+			ret = dump_ipseckey(&p);
 			break;
 		case KNOT_RRTYPE_RRSIG:
-			ret = dump_rrsig(data, data_len, dst, maxlen, style);
+			ret = dump_rrsig(&p);
 			break;
 		case KNOT_RRTYPE_NSEC:
-			ret = dump_nsec(data, data_len, dst, maxlen, style);
+			ret = dump_nsec(&p);
 			break;
 		case KNOT_RRTYPE_DHCID:
-			ret = dump_dhcid(data, data_len, dst, maxlen, style);
+			ret = dump_dhcid(&p);
 			break;
 		case KNOT_RRTYPE_NSEC3:
-			ret = dump_nsec3(data, data_len, dst, maxlen, style);
+			ret = dump_nsec3(&p);
 			break;
 		case KNOT_RRTYPE_NSEC3PARAM:
-			ret = dump_nsec3param(data, data_len, dst, maxlen, style);
+			ret = dump_nsec3param(&p);
 			break;
 		case KNOT_RRTYPE_TLSA:
-			ret = dump_tlsa(data, data_len, dst, maxlen, style);
+			ret = dump_tlsa(&p);
 			break;
 		case KNOT_RRTYPE_NID:
 		case KNOT_RRTYPE_L64:
-			ret = dump_l64(data, data_len, dst, maxlen, style);
+			ret = dump_l64(&p);
 			break;
 		case KNOT_RRTYPE_L32:
-			ret = dump_l32(data, data_len, dst, maxlen, style);
+			ret = dump_l32(&p);
 			break;
 		case KNOT_RRTYPE_EUI48:
 		case KNOT_RRTYPE_EUI64:
-			ret = dump_eui(data, data_len, dst, maxlen, style);
+			ret = dump_eui(&p);
 			break;
 		case KNOT_RRTYPE_TSIG:
-			ret = dump_tsig(data, data_len, dst, maxlen, style);
+			ret = dump_tsig(&p);
 			break;
 		default:
-			ret = dump_unknown(data, data_len, dst, maxlen, style);
+			ret = dump_unknown(&p);
 			break;
 	}
 
@@ -1906,6 +1826,7 @@ int knot_rrset_txt_dump_data(const knot_rrset_t      *rrset,
 	}
 
 int knot_rrset_txt_dump_header(const knot_rrset_t      *rrset,
+                               const uint32_t          ttl,
                                char                    *dst,
                                const size_t            maxlen,
                                const knot_dump_style_t *style)
@@ -1919,7 +1840,7 @@ int knot_rrset_txt_dump_header(const knot_rrset_t      *rrset,
 	int    ret;
 
 	// Dump rrset owner.
-	char *name = knot_dname_to_str(rrset->owner);
+	char *name = knot_dname_to_str_alloc(rrset->owner);
 	if (style->ascii_to_idn != NULL) {
 		style->ascii_to_idn(&name);
 	}
@@ -1934,16 +1855,17 @@ int knot_rrset_txt_dump_header(const knot_rrset_t      *rrset,
 
 	// Dump rrset ttl.
 	if (style->show_ttl) {
-		if (style->human_ttl) {
+		if (style->empty_ttl) {
+			ret = snprintf(dst + len, maxlen - len, "%c", sep);
+		} else if (style->human_ttl) {
 			// Create human readable ttl string.
-			if (time_to_human_str(buf, sizeof(buf), rrset->ttl) < 0) {
+			if (time_to_human_str(buf, sizeof(buf), ttl) < 0) {
 				return KNOT_ESPACE;
 			}
 			ret = snprintf(dst + len, maxlen - len, "%s%c",
 			               buf, sep);
 		} else {
-			ret = snprintf(dst + len, maxlen - len, "%u%c",
-			               rrset->ttl, sep);
+			ret = snprintf(dst + len, maxlen - len, "%u%c", ttl, sep);
 		}
 		SNPRINTF_CHECK(ret, maxlen - len);
 		len += ret;
@@ -1963,7 +1885,7 @@ int knot_rrset_txt_dump_header(const knot_rrset_t      *rrset,
 	if (knot_rrtype_to_string(rrset->type, buf, sizeof(buf)) < 0) {
 		return KNOT_ESPACE;
 	}
-	if (rrset->rdata_count > 0) {
+	if (rrset->rrs.rr_count > 0) {
 		ret = snprintf(dst + len, maxlen - len, "%s%c", buf, sep);
 	} else {
 		ret = snprintf(dst + len, maxlen - len, "%s", buf);
@@ -1977,8 +1899,6 @@ int knot_rrset_txt_dump_header(const knot_rrset_t      *rrset,
 int knot_rrset_txt_dump(const knot_rrset_t      *rrset,
                         char                    *dst,
                         const size_t            maxlen,
-                        const bool              dump_rdata,
-                        const bool              dump_rrsig,
                         const knot_dump_style_t *style)
 {
 	if (rrset == NULL || dst == NULL || style == NULL) {
@@ -1988,60 +1908,32 @@ int knot_rrset_txt_dump(const knot_rrset_t      *rrset,
 	size_t len = 0;
 	int    ret;
 
-	if (dump_rdata) {
-		// APL RR or UPDATE delete may have empty RDATA => dump header.
-		if (rrset->rdata_count == 0) {
-			// Dump rdata owner, class, ttl and type.
-			ret = knot_rrset_txt_dump_header(rrset, dst + len,
-			                                 maxlen - len, style);
-			if (ret < 0) {
-				return KNOT_ESPACE;
-			}
-			len += ret;
-
-			// Terminate line.
-			if (len >= maxlen) {
-				return KNOT_ESPACE;
-			}
-			dst[len++] = '\n';
-			dst[len] = '\0';
-		}
-
-		// Loop over rdata in rrset.
-		for (size_t i = 0; i < rrset->rdata_count; i++) {
-			// Dump rdata owner, class, ttl and type.
-			ret = knot_rrset_txt_dump_header(rrset, dst + len,
-			                                 maxlen - len, style);
-			if (ret < 0) {
-				return KNOT_ESPACE;
-			}
-			len += ret;
-
-			// Dump rdata as such.
-			ret = knot_rrset_txt_dump_data(rrset, i, dst + len,
-			                               maxlen - len, style);
-			if (ret < 0) {
-				return KNOT_ESPACE;
-			}
-			len += ret;
-
-			// Terminate line.
-			if (len >= maxlen) {
-				return KNOT_ESPACE;
-			}
-			dst[len++] = '\n';
-			dst[len] = '\0';
-		}
-	}
-
-	// Dump RRSIG records if any via recursion call.
-	if (dump_rrsig && rrset->rrsigs != NULL) {
-		ret = knot_rrset_txt_dump(rrset->rrsigs, dst + len,
-		                          maxlen - len, true, false, style);
+	// Loop over rdata in rrset.
+	uint16_t rr_count = rrset->rrs.rr_count;
+	for (uint16_t i = 0; i < rr_count; i++) {
+		// Dump rdata owner, class, ttl and type.
+		const knot_rdata_t *rr_data = knot_rdataset_at(&rrset->rrs, i);
+		ret = knot_rrset_txt_dump_header(rrset, knot_rdata_ttl(rr_data),
+		                                 dst + len, maxlen - len, style);
 		if (ret < 0) {
 			return KNOT_ESPACE;
 		}
 		len += ret;
+
+		// Dump rdata as such.
+		ret = knot_rrset_txt_dump_data(rrset, i, dst + len,
+		                               maxlen - len, style);
+		if (ret < 0) {
+			return KNOT_ESPACE;
+		}
+		len += ret;
+
+		// Terminate line.
+		if (len >= maxlen) {
+			return KNOT_ESPACE;
+		}
+		dst[len++] = '\n';
+		dst[len] = '\0';
 	}
 
 	return len;
